from abc import ABC
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    AsyncExitStack,
)
from typing import Any, Callable, ClassVar, Dict, Type, TypeVar, cast, overload

from imbue.abc import InternalContainer
from imbue.contexts.base import Context, ContextualizedProvider
from imbue.dependency import Interface
from imbue.providers.instance import DelegatedInstanceProvider

V = TypeVar("V")


class ContextualizedContainer(AsyncExitStack, ABC):
    """Wraps the container to support context handling.
    This will be responsible for storing already provided dependencies for a particular context.
    """

    CONTEXT: ClassVar[Context]

    def __init__(
        self,
        container: InternalContainer,
        contextualized: Dict[Context, "ContextualizedContainer"],
    ):
        super().__init__()
        self._container = container
        self._contextualized = dict(contextualized)
        self._contextualized[self.CONTEXT] = self
        self._provided: Dict[Interface, Any] = {}

    @overload
    async def get(self, interface: Type[V]) -> V:
        """Specific type annotation for classes."""

    @overload
    async def get(self, interface: Callable) -> Callable:
        """Specific type annotation for functions."""

    async def get(self, interface: Interface) -> Any:
        """Find the proper container based on context and provide."""
        provider = self._container.get_provider(interface)
        return await self._contextualized[
            cast(Context, provider.context)
        ]._get_or_provide(provider)

    async def _get_or_provide(self, provider: ContextualizedProvider) -> Any:
        """Get from already provided or provide the dependency."""
        if provided := self._provided.get(provider.interface):
            return provided
        provided = await self._provide(provider)
        self._provided[provider.interface] = provided
        return provided

    async def _provide(self, provider: ContextualizedProvider) -> Any:
        """Actually provide the dependency."""
        provided = await provider.get(
            **{
                s.name: await self.get(s.interface)
                for s in self._container.get_sub_dependencies(provider.interface)
            },
        )
        if (
            isinstance(provider.provider, DelegatedInstanceProvider)
            and provider.provider.is_context_manager
        ):
            if isinstance(provided, AbstractAsyncContextManager):
                return await self.enter_async_context(provided)
            if isinstance(provided, AbstractContextManager):
                return self.enter_context(provided)
        return provided

    async def init(self) -> None:
        """Init eager dependencies."""
        for provider in self._container.get_eager_providers(self.CONTEXT):
            await self._get_or_provide(provider)

    async def __aenter__(self):
        await self.init()
        return self

    async def close(self) -> None:
        # No exception here.
        await self.__aexit__(None, None, None)



from typing import Any, AsyncIterator

from fastapi import Request
from fastapi.params import Depends
from typing_extensions import Annotated

from imbue.contexts.task import TaskContainer
from imbue.dependency import Interface


async def _task_container(request: Request) -> AsyncIterator[TaskContainer]:
    """Initialize the task container.
    This needs to be done for every request.
    For more details, see:
        - https://fastapi.tiangolo.com/tutorial/dependencies/global-dependencies/
        - https://fastapi.tiangolo.com/tutorial/dependencies/
    """
    async with request.state.app_container.task_context() as container:
        yield container


request_lifespan = Depends(_task_container, use_cache=True)


class Dependency(Depends):
    def __init__(self, interface: Interface):
        async def _get(container: Annotated[TaskContainer, request_lifespan]) -> Any:
            return await container.get(interface)

        super().__init__(_get, use_cache=False)



from typing import Any

from imbue.contexts.abc import ContextualizedContainer
from imbue.contexts.base import Context, ContextualizedProvider, make_context_decorator

factory_context = make_context_decorator(Context.FACTORY)


class FactoryContainer(ContextualizedContainer):
    CONTEXT = Context.FACTORY

    async def _get_or_provide(self, provider: ContextualizedProvider) -> Any:
        """Always provide."""
        return await self._provide(provider)



import inspect
from typing import ClassVar, Iterable, Iterator, Union

from imbue.contexts.base import (
    ContextualizedDependency,
    ContextualizedProvider,
    DelegatedProviderWrapper,
)
from imbue.dependency import Dependency


class Package:
    """Container class to group dependency providers.
    Each dependency is provided by a method.
    Additionally, you can specify dependencies for which no
    provider function is required in `EXTRA_DEPENDENCIES`.
    """

    EXTRA_DEPENDENCIES: ClassVar[
        Iterable[Union[Dependency, ContextualizedDependency]]
    ] = ()

    def get_providers(self) -> Iterator[ContextualizedProvider]:
        for dependency in self.EXTRA_DEPENDENCIES:
            if isinstance(dependency, ContextualizedDependency):
                yield from dependency.get_providers()
            else:
                yield from ContextualizedProvider.from_dependency(dependency)
        for _, member in inspect.getmembers(self):
            if not isinstance(member, DelegatedProviderWrapper):
                continue
            # Now that we have an instance to bind, we can get the final provider.
            yield member.to_contextualized_provider(instance=self)



import asyncio
from typing import Any, AsyncContextManager, Callable, Dict, Type, overload

from imbue.abc import InternalContainer
from imbue.contexts.abc import ContextualizedContainer, V
from imbue.contexts.base import Context, make_context_decorator
from imbue.contexts.task import TaskContainer
from imbue.dependency import Interface

thread_context = make_context_decorator(Context.THREAD)


class ThreadContainer(ContextualizedContainer):
    CONTEXT = Context.THREAD

    def __init__(
        self,
        container: InternalContainer,
        contextualized: Dict[Context, "ContextualizedContainer"],
    ):
        super().__init__(container, contextualized)
        self._locks: Dict[Interface, AsyncContextManager] = {}

    @overload
    async def get(self, interface: Type[V]) -> V:
        """Specific type annotation for classes."""

    @overload
    async def get(self, interface: Callable) -> Callable:
        """Specific type annotation for functions."""

    async def get(self, interface: Interface) -> Any:
        if provided := self._provided.get(interface):
            return provided
        if interface not in self._locks:
            self._locks[interface] = asyncio.Lock()
        async with self._locks[interface]:
            return await super().get(interface)

    def task_context(self) -> "TaskContainer":
        """Spawn registries for each task."""
        return TaskContainer(self._container, self._contextualized)



from typing import Any, Callable, Iterator, Type

from imbue.dependency import SubDependency
from imbue.providers.abc import Provider
from imbue.utils import get_annotations, partial


class FunctionProvider(Provider[Callable, Callable]):
    """Automatically enrich function arguments with dependencies."""

    @property
    def sub_dependencies(self) -> Iterator[SubDependency]:
        for name, annotation in get_annotations(
            self.interface,
            with_return=False,
        ).items():
            # Make all dependencies optional for functions,
            # this will allow to pass other arguments in functions being injected.
            # This will be equivalent to using partial with dependencies already passed.
            yield SubDependency(name, annotation.annotation, mandatory=False)

    async def get(self, **dependencies: Any) -> Callable:
        return partial(self.interface, **dependencies)


class MethodProvider(Provider[Callable, Callable]):
    """The class is instantiated with dependencies and the bound method is returned."""

    def __init__(self, func: Callable, cls: Type):
        super().__init__(func)
        self._cls = cls

    @property
    def sub_dependencies(self) -> Iterator[SubDependency]:
        # The real dependencies are handled in the instance.
        yield SubDependency("__instance__", self._cls)
        for name, annotation in get_annotations(
            self.interface,
            with_return=False,
            with_instance=False,
        ).items():
            # Make all dependencies optional for functions,
            # this will allow to pass other arguments in functions being injected.
            # This will be equivalent to using partial with dependencies already passed.
            yield SubDependency(name, annotation.annotation, mandatory=False)

    async def get(self, **dependencies: Any) -> Callable:
        instance = dependencies.pop("__instance__")
        return partial(getattr(instance, self.interface.__name__), **dependencies)



from dataclasses import dataclass
from typing import Callable, Generic, Type, TypeVar, Union

Interface = Union[Type, Callable]


@dataclass
class SubDependency:
    """Internally specify sub dependencies."""

    # The name of the dependency argument.
    name: str
    # The external interface of the dependency.
    interface: Interface
    # The dependency is required or optional.
    # If required it will have to be in the container.
    # If optional it will only be provided if in the container.
    # This means that the sub dependency needs a default value provided.
    # This is explicitly not named `optional` to avoid confusion with
    # `Optional` type annotation that means nullable.
    mandatory: bool = True


T = TypeVar("T", bound=Interface)


@dataclass
class Interfaced(Generic[T]):
    """Allows defining dependencies around interfaces.

    The exposed type will be the interface.
    The sub dependencies will the inferred from the implementation.
    The injected dependency will be the instantiated implementation.
    """

    # The external interface of the dependency.
    interface: T
    # The internal implementation of the interface.
    implementation: T


Dependency = Union[Interface, Interfaced]



from contextlib import asynccontextmanager
from typing import Any, AsyncContextManager, AsyncIterator, Callable, TypedDict

from imbue.container import Container
from imbue.contexts.application import ApplicationContainer


class State(TypedDict):
    app_container: ApplicationContainer


def app_lifespan(container: Container) -> Callable[[Any], AsyncContextManager[State]]:
    @asynccontextmanager
    async def _lifespan(_) -> AsyncIterator[State]:
        """Initializes the application container.
        This should be passed to the `lifespan` init parameter of FastAPI.
        For more details, see
            - https://fastapi.tiangolo.com/advanced/events
            - https://www.starlette.io/lifespan/.
        """
        async with container.application_context() as app_container:
            yield {"app_container": app_container}

    return _lifespan



from imbue.contexts.abc import ContextualizedContainer
from imbue.contexts.base import Context, make_context_decorator
from imbue.contexts.factory import FactoryContainer

task_context = make_context_decorator(Context.TASK)


class TaskContainer(ContextualizedContainer):
    CONTEXT = Context.TASK

    async def init(self) -> None:
        await super().init()
        # Init the factory container.
        container = FactoryContainer(self._container, self._contextualized)
        self._contextualized[container.CONTEXT] = container
        await self.enter_async_context(container)



import functools
import inspect
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Set,
    get_type_hints,
)

from imbue.exceptions import DependencyError


@dataclass
class Annotation:
    annotation: Any
    mandatory: bool


def get_annotations(
    func: Callable,
    with_return: bool = True,
    with_instance: bool = True,
) -> Dict[str, Annotation]:
    """Wrapper around signature and get_type_hints functions.
    Note: variadic and positional only parameters are excluded as those make injection risky.
    """
    hints = get_type_hints(func)
    signature = inspect.signature(func)
    annotations = {
        # Use annotations or default to signature annotation.
        # We still have to use the annotations since they are resolved if needed ("Cls" -> Cls).
        p.name: Annotation(
            hints[p.name] if p.name in hints else p.annotation,
            mandatory=p.default is p.empty,
        )
        for p in signature.parameters.values()
        if p.kind
        not in (
            inspect.Parameter.VAR_POSITIONAL,
            inspect.Parameter.VAR_KEYWORD,
            inspect.Parameter.VAR_POSITIONAL,
        )
        and (with_instance or p.name not in ("self", "cls"))
    }
    if with_return:
        annotations["return"] = Annotation(
            hints["return"] if "return" in hints else signature.return_annotation,
            mandatory=False,
        )
    return annotations


def partial(func: Callable, **kwargs: Any) -> Callable:
    """Replacement of `functools.partial` to make it work with type hints.
    Note: this is just smoke to allow programmatically parsing the signature.
    """
    if inspect.iscoroutinefunction(func):

        @functools.wraps(func)
        async def _wrapper(*a, **kw):
            return await func(*a, **kwargs, **kw)

    else:

        @functools.wraps(func)
        def _wrapper(*a, **kw):
            return func(*a, **kwargs, **kw)

    # Update the annotations and signature of the function to remove injected arguments.
    _wrapper.__annotations__ = {
        k: v.annotation for k, v in get_annotations(func).items() if k not in kwargs
    }
    signature = inspect.signature(func)
    _wrapper.__signature__ = signature.replace(  # type: ignore[attr-defined]
        parameters=[p for p in signature.parameters.values() if p.name not in kwargs],
    )
    return _wrapper


def extend(
    func: Callable,
    remove_instance: bool = False,
    remove: Optional[Set[str]] = None,
) -> Callable:
    """Decorator to extend function in order to add parameters and fix signature.
    Note: this is just smoke to allow programmatically parsing the signature.
    Parameters should be added first, and have a name with no risk to collide with existing parameters.
    """
    remove_params: Set[str] = remove or set()
    if remove_instance:
        remove_params.add("self")

    def _wrapper(wrapped: Callable) -> Callable:
        wrapped_hints = get_annotations(wrapped, with_return=False)
        _wrapped = functools.wraps(func)(wrapped)

        # Handle annotations.
        # Shallow copy to avoid altering original function.
        _wrapped.__annotations__ = dict(_wrapped.__annotations__)
        for name, ann in wrapped_hints.items():
            if name in _wrapped.__annotations__:
                raise DependencyError(
                    f"parameter {name} already declared in function {func!r}",
                )
            _wrapped.__annotations__[name] = ann.annotation
        for name in remove_params:
            _wrapped.__annotations__.pop(name, None)

        signature = inspect.signature(func)
        _wrapped.__signature__ = signature.replace(  # type: ignore[attr-defined]
            parameters=[
                inspect.Parameter(
                    n,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=a.annotation,
                )
                for n, a in wrapped_hints.items()
            ]
            + [p for p in signature.parameters.values() if p.name not in remove_params],
        )
        return _wrapped

    return _wrapper



class DependencyError(Exception):
    """Generic dependency error."""


class UnsupportedDependencyInterfaceError(DependencyError):
    """The interface is not supported by this library."""


class DependencyResolutionError(DependencyError):
    """Dependencies could not be resolved."""



import inspect
from typing import Iterator

from imbue.dependency import Dependency, Interfaced
from imbue.exceptions import UnsupportedDependencyInterfaceError
from imbue.providers.abc import Provider
from imbue.providers.function import (
    FunctionProvider,
    MethodProvider,
)
from imbue.providers.instance import (
    InstanceProvider,
    InterfacedInstanceProvider,
)


def get_providers(dependency: Dependency) -> Iterator[Provider]:
    """Get the providers or raise if incorect type."""
    if isinstance(dependency, Interfaced):
        yield InterfacedInstanceProvider(dependency)
    elif inspect.isclass(dependency):
        yield InstanceProvider(dependency)
    elif inspect.iscoroutinefunction(dependency) or inspect.isfunction(dependency):
        # Determine if it's a method of function.
        func_names = dependency.__qualname__.split(".")
        if len(func_names) == 2:
            cls = getattr(
                inspect.getmodule(dependency),
                func_names[0],
                None,
            )
            if cls is None:
                raise UnsupportedDependencyInterfaceError(
                    f"could not find class for function {dependency!r}",
                )
            # Also yield the instance provider.
            yield InstanceProvider(cls)
            yield MethodProvider(dependency, cls)
        elif len(func_names) == 1:
            yield FunctionProvider(dependency)
        else:
            raise UnsupportedDependencyInterfaceError(
                f"{dependency!r} has too many parts in qualname",
            )
    else:
        raise UnsupportedDependencyInterfaceError(
            f"{dependency!r} must be a class, function or method",
        )



from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Iterator, List, Union, cast

from imbue.abc import InternalContainer
from imbue.contexts.application import ApplicationContainer
from imbue.contexts.base import (
    Context,
    ContextualizedDependency,
    ContextualizedProvider,
)
from imbue.dependency import Dependency, Interface, SubDependency
from imbue.exceptions import DependencyResolutionError
from imbue.package import Package


@dataclass
class DependencyChain:
    """Dependency chain, representing the graph of dependencies.
    Its role is to check whether there are no cycles nor context issues.
    """

    chain: List[ContextualizedProvider]

    def add(self, provider: ContextualizedProvider) -> "DependencyChain":
        """Create a new chain, adding the provider at the end."""
        chain = DependencyChain([*self.chain, provider])
        # Check for cycles.
        if provider in self.chain:
            raise DependencyResolutionError(f"circular dependency found:\n{chain}")
        return chain

    def check(self) -> None:
        if self.last.context is None:
            raise DependencyResolutionError(
                f"provider {self.last} does not have a context set"
            )
        # If previous context is not set, it will be done automatically so no check is required.
        if len(self.chain) < 2 or self.chain[-2].context is None:
            return
        # The deeper the chain, the lower the context must be.
        # App dependencies cannot have task dependencies but the inverse is possible.
        if self.last.context > self.chain[-2].context:
            raise DependencyResolutionError(f"context error:\n{self}")

    @property
    def last(self) -> ContextualizedProvider:
        """Get the last provider in the chain."""
        return self.chain[-1]

    def __str__(self) -> str:
        return "\n".join(
            (
                f"{' ' * i}-> {p.interface} ({p.context})"
                for i, p in enumerate(self.chain)
            ),
        )


class Container(InternalContainer):
    def __init__(
        self,
        *dependencies_or_packages: Union[Dependency, ContextualizedDependency, Package],
    ):
        # The link between an interface and its provider.
        self._providers: Dict[Interface, ContextualizedProvider] = {}
        # Cache sub dependencies for each interface.
        self._sub_dependencies: Dict[Interface, List[SubDependency]] = {}
        # All providers that should be eager inited.
        self._by_context_eager_providers: Dict[
            Context, List[ContextualizedProvider]
        ] = defaultdict(
            list,
        )

        # Add all dependencies.
        for dep_or_pkg in dependencies_or_packages:
            if isinstance(dep_or_pkg, (ContextualizedDependency, Package)):
                providers_iterator = dep_or_pkg.get_providers()
            else:
                providers_iterator = ContextualizedProvider.from_dependency(dep_or_pkg)
            for provider in providers_iterator:
                if provider.interface in self._providers:
                    raise DependencyResolutionError(
                        "multiple providers found for the same type: "
                        f"{self._providers[provider.interface]!r}, {provider!r}",
                    )
                self._providers[provider.interface] = provider
        # Resolve the graph.
        for provider in self._providers.values():
            self._resolve(DependencyChain([provider]))

    def _resolve(self, chain: DependencyChain) -> None:
        """Construct the graph of sub dependencies."""
        provider = chain.last
        if provider.interface in self._sub_dependencies:
            # Already handled, we just need to check the full chain.
            chain.check()
            return
        dependencies: List[SubDependency] = []
        sub_providers: List[ContextualizedProvider] = []
        for sub_dependency in provider.sub_dependencies:
            if (
                not sub_dependency.mandatory
                and sub_dependency.interface not in self._providers
            ):
                continue
            if sub_dependency.interface not in self._providers:
                raise DependencyResolutionError(
                    f"no provider found for {sub_dependency.interface}, from provider {provider!r}",
                )
            sub_provider = self._providers[sub_dependency.interface]
            sub_providers.append(sub_provider)
            self._resolve(chain.add(sub_provider))
            dependencies.append(sub_dependency)
        # Set the context automatically based on dependencies if not set.
        # We want to set the lowest context possible.
        if provider.context is None:
            provider.context = (
                max(cast(Context, s.context) for s in sub_providers)
                if sub_providers
                else Context.APPLICATION
            )
        chain.check()
        self._sub_dependencies[provider.interface] = dependencies
        if provider.eager:
            self._by_context_eager_providers[provider.context].append(provider)

    def add(self, dependency: Dependency, context: Context = Context.TASK) -> None:
        """Add another interface, used to eagerly add all task functions/methods as providers.
        This allows to make all necessary checks at application start rather than during task processing.
        """
        for provider in ContextualizedProvider.from_dependency(
            dependency=dependency,
            context=context,
        ):
            if provider.interface in self._providers:
                continue
            self._providers[provider.interface] = provider
            self._resolve(DependencyChain([provider]))

    def get_provider(self, interface: Interface) -> ContextualizedProvider:
        """Get the provider for an interface."""
        if interface not in self._providers:
            raise DependencyResolutionError(f"unknow interface {interface}")
        return self._providers[interface]

    def get_sub_dependencies(self, interface: Interface) -> Iterator[SubDependency]:
        """Get all sub dependencies for an interface."""
        yield from self._sub_dependencies[interface]

    def get_eager_providers(self, context: Context) -> Iterator[ContextualizedProvider]:
        """Get all providers that should be eager inited for a context."""
        return iter(self._by_context_eager_providers[context])

    def application_context(self) -> "ApplicationContainer":
        """Spawns the first contextualized container on the application level."""
        return ApplicationContainer(self, {})



from __future__ import annotations

import inspect
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from enum import IntEnum
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    Callable,
    ContextManager,
    Generic,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from imbue.dependency import Dependency, SubDependency
from imbue.providers.abc import Provider
from imbue.providers.common import get_providers
from imbue.providers.instance import DelegatedInstanceProvider
from imbue.utils import partial


class Context(IntEnum):
    """Supported contexts, lower have longer lifetime."""

    APPLICATION = 10  # Equivalent to singletons.
    THREAD = 20
    TASK = 30
    FACTORY = 40  # Never reused.


T = TypeVar("T")
V = TypeVar("V")


@dataclass
class ContextualizedDependency:
    dependency: Dependency
    context: Optional[Context] = None
    eager: bool = False

    def get_providers(self) -> Iterator[ContextualizedProvider]:
        yield from ContextualizedProvider.from_dependency(
            self.dependency,
            self.context,
            self.eager,
        )


@dataclass
class ContextualizedProvider(Generic[T, V]):
    """Wrap a provider to handle context and lifetime."""

    provider: Provider[T, V]
    context: Optional[Context]
    eager: bool

    @classmethod
    def from_dependency(
        cls,
        dependency: Dependency,
        context: Optional[Context] = None,
        eager: bool = False,
    ) -> Iterator["ContextualizedProvider"]:
        """In some cases, an interface yields multiple providers.
        Ex: a method yields a provider for a class and one for the method.
        """
        for provider in get_providers(dependency):
            yield cls(
                provider=provider,
                context=context,
                eager=eager,
            )

    @property
    def interface(self) -> T:
        return self.provider.interface

    @property
    def sub_dependencies(self) -> Iterator[SubDependency]:
        yield from self.provider.sub_dependencies

    async def get(
        self,
        **dependencies: Any,
    ) -> Union[V, ContextManager[V], AsyncContextManager[V]]:
        return await self.provider.get(**dependencies)


@dataclass
class DelegatedProviderWrapper(Generic[V]):
    """Wrapper for delegated provider methods.
    Its purpose is to wait until we have an instance to pass to the method.
    """

    func: Callable[..., Union[V, Iterator[V], AsyncIterator[V]]]
    context: Optional[Context]
    eager: bool

    def to_contextualized_provider(
        self,
        instance: Any,
    ) -> ContextualizedProvider[Type[V], V]:
        """Get the provider from this wrapper."""
        func, is_context_manager = self._get_func()
        return ContextualizedProvider(
            provider=DelegatedInstanceProvider(
                provider_func=partial(func, self=instance),
                is_context_manager=is_context_manager,
            ),
            context=self.context,
            eager=self.eager,
        )

    def _get_func(
        self,
    ) -> Tuple[
        Callable[
            ...,
            Union[V, ContextManager[V], AsyncContextManager[V]],
        ],
        bool,
    ]:
        """Support for context managers."""
        if inspect.isasyncgenfunction(self.func):
            return asynccontextmanager(self.func), True
        if inspect.isgeneratorfunction(self.func):
            return contextmanager(self.func), True
        return self.func, False  # type: ignore[return-value]


def make_context_decorator(context: Optional[Context]):
    """Wrap a delegated function providing an interface to assign a context and handle eagerness."""

    def _wrapper(func: Optional[Callable] = None, *, eager: bool = False):
        def wrap(fn: Callable) -> DelegatedProviderWrapper:
            return DelegatedProviderWrapper(func=fn, context=context, eager=eager)

        # Check if called like `@context` or `@context()`.
        if func is None:
            # Called with parentheses.
            return wrap
        # Called without parentheses.
        return wrap(func)

    return _wrapper


auto_context = make_context_decorator(None)



from abc import ABC, abstractmethod
from typing import Iterator

from imbue.contexts.base import Context, ContextualizedProvider
from imbue.dependency import Interface, SubDependency


class InternalContainer(ABC):
    """Internal abstract to define an interface to a container to other parts of the injection system."""

    @abstractmethod
    def get_provider(self, interface: Interface) -> ContextualizedProvider:
        """Get a provider for an interface."""

    @abstractmethod
    def get_sub_dependencies(self, interface: Interface) -> Iterator[SubDependency]:
        """Get all dependencies from a provider."""

    @abstractmethod
    def get_eager_providers(self, context: Context) -> Iterator[ContextualizedProvider]:
        """Get all eager providers for a context."""



from abc import ABC, abstractmethod
from typing import (
    Any,
    AsyncContextManager,
    ContextManager,
    Generic,
    Iterator,
    TypeVar,
    Union,
)

from imbue.dependency import SubDependency

T = TypeVar("T")
V = TypeVar("V")


class Provider(Generic[T, V], ABC):
    """The foundation of dependency injection.
    The role of the provider is to expose sub dependencies and provide the dependencies given sub dependencies.
    """

    def __init__(self, interface: T):
        self.interface: T = interface

    @property
    @abstractmethod
    def sub_dependencies(self) -> Iterator[SubDependency]:
        """Get the dependencies from the interface."""

    @abstractmethod
    async def get(
        self,
        **dependencies: Any,
    ) -> Union[V, ContextManager[V], AsyncContextManager[V]]:
        """Provide the dependency for the interface."""

    def __repr__(self) -> str:
        return f"{type(self)}(interface={self.interface})"



from imbue.container import Container
from imbue.contexts.application import (
    ApplicationContainer,
    application_context,
)
from imbue.contexts.base import (
    Context,
    ContextualizedDependency,
    ContextualizedProvider,
    auto_context,
)
from imbue.contexts.factory import FactoryContainer, factory_context
from imbue.contexts.task import TaskContainer, task_context
from imbue.contexts.thread import ThreadContainer, thread_context
from imbue.dependency import Interfaced
from imbue.package import Package
from imbue.utils import extend, get_annotations, partial



import inspect
from typing import (
    Any,
    AsyncContextManager,
    Callable,
    ContextManager,
    Generic,
    Iterator,
    Type,
    TypeVar,
    Union,
    get_args,
)

from imbue.dependency import Interfaced, SubDependency
from imbue.providers.abc import Provider
from imbue.utils import get_annotations

C = TypeVar("C")


class InstanceProvider(Provider[Type[C], C], Generic[C]):
    """Create instances, filling in dependencies."""

    @property
    def sub_dependencies(self) -> Iterator[SubDependency]:
        for name, annotation in get_annotations(
            self.interface.__init__,
            with_return=False,
            with_instance=False,
        ).items():
            yield SubDependency(name, annotation.annotation, annotation.mandatory)

    async def get(self, **dependencies: Any) -> C:
        return self.interface(**dependencies)


class InterfacedInstanceProvider(Provider[Type[C], C], Generic[C]):
    """Create instances, using an interface as the dependency type, filling in dependencies."""

    def __init__(self, dependency: Interfaced[Type[C]]):
        super().__init__(dependency.interface)
        self.implementation: Type[C] = dependency.implementation

    @property
    def sub_dependencies(self) -> Iterator[SubDependency]:
        for name, annotation in get_annotations(
            self.implementation.__init__,
            with_return=False,
            with_instance=False,
        ).items():
            yield SubDependency(name, annotation.annotation, annotation.mandatory)

    async def get(self, **dependencies: Any) -> C:
        return self.implementation(**dependencies)


class DelegatedInstanceProvider(Provider[Type[C], C], Generic[C]):
    """Create instances, delegating creation to a function."""

    def __init__(
        self,
        provider_func: Callable[
            ...,
            Union[C, ContextManager[C], AsyncContextManager[C]],
        ],
        is_context_manager: bool,
    ):
        self._provider_func = provider_func
        self._awaitable: bool = False
        self.is_context_manager = is_context_manager

        # Get the proper return type and provider func based on different cases.
        # In case it's a generator, wrap in a context manager and get the underlying return type.
        return_annotation: Type[C]
        if is_context_manager:
            return_annotation, *_ = get_args(
                get_annotations(provider_func)["return"].annotation
            )
        else:
            return_annotation = get_annotations(provider_func)["return"].annotation
            if inspect.iscoroutinefunction(provider_func):
                self._awaitable = True
        super().__init__(return_annotation)

    @property
    def sub_dependencies(self) -> Iterator[SubDependency]:
        for name, annotation in get_annotations(
            self._provider_func,
            with_return=False,
            with_instance=False,
        ).items():
            yield SubDependency(name, annotation.annotation, annotation.mandatory)

    async def get(
        self,
        **dependencies: Any,
    ) -> Union[C, ContextManager[C], AsyncContextManager[C]]:
        if self._awaitable:
            return await self._provider_func(**dependencies)  # type: ignore[misc]
        return self._provider_func(**dependencies)
